import parser from 'postcss-selector-parser';
import Container from 'postcss/lib/container';

const NTH_CHILD_QUERY_REGEX = /^(-)?(\d*)?n$/; // eslint-disable-line unicorn/no-unsafe-regex

function normalizeTagSelector(selectors) {
	selectors.walk((selector, index) => {
		if (['selector', 'tag', 'universal'].includes(selector.type)) {
			return;
		}

		let referenceSelector = null;

		if (index === 0 && selector.type !== 'combinator') {
			referenceSelector = selector;
		}
		if (index !== 0 && selector.type === 'combinator') {
			referenceSelector = selector.next();
		}

		if (
			referenceSelector !== null &&
			!['tag', 'universal'].includes(referenceSelector.type)
		) {
			selector.parent.insertBefore(referenceSelector, parser.universal());
		}
	});
}

function normalizeNthChildSelector(selectors) {
	selectors.walkTags((number) => {
		if (number.parent.parent && number.parent.parent.type !== 'pseudo') {
			return;
		}

		const position = Number(number.value);
		const previousSelector = number.prev();
		const nextSelector = number.next();

		if (
			number.type === 'tag' &&
			Number.isInteger(position) &&
			position >= 0 &&
			!previousSelector &&
			!nextSelector
		) {
			number.parent.insertBefore(number, parser.tag({ value: '0n' }));
			number.parent.insertBefore(
				number,
				parser.combinator({ value: '+' })
			);
			number.parent.insertBefore(number, parser.tag({ value: position }));
			number.remove();
		}

		if (number.type === 'tag' && number.value === 'odd') {
			number.parent.insertBefore(number, parser.tag({ value: '2n' }));
			number.parent.insertBefore(
				number,
				parser.combinator({ value: '+' })
			);
			number.parent.insertBefore(number, parser.tag({ value: '1' }));
			number.remove();
		}

		if (number.type === 'tag' && number.value === 'even') {
			number.parent.insertBefore(number, parser.tag({ value: '2n' }));
			number.parent.insertBefore(
				number,
				parser.combinator({ value: '+' })
			);
			number.parent.insertBefore(number, parser.tag({ value: '2' }));
			number.remove();
		}

		if (NTH_CHILD_QUERY_REGEX.test(number.value)) {
			const [, negative = '', regexNumber = 1] = number.value.match(
				NTH_CHILD_QUERY_REGEX
			);
			const isAlone = !nextSelector;
			const isInvalidCombinator =
				nextSelector &&
				nextSelector.type !== 'combinator' &&
				nextSelector.value !== '+';

			if (isInvalidCombinator) {
				number.parent.insertBefore(number, parser.tag({ value: '-n' }));
				number.parent.insertBefore(
					number,
					parser.combinator({ value: '+' })
				);
				number.parent.insertBefore(number, parser.tag({ value: 0 }));
			} else {
				const value = isAlone
					? `${regexNumber}n`
					: `${negative}${regexNumber}n`;

				number.parent.insertBefore(number, parser.tag({ value }));

				if (isAlone && negative === '') {
					number.parent.insertBefore(
						number,
						parser.combinator({ value: '+' })
					);
					number.parent.insertBefore(
						number,
						parser.tag({ value: regexNumber })
					);
				}
			}

			number.remove();
		}
	});
}

var getSelectorAst = (query) =>
	parser((selectors) => {
		normalizeTagSelector(selectors);
		normalizeNthChildSelector(selectors);
	}).ast(query);

class PassthroughContainer extends Container {
	constructor() {
		super();
		this.nodes = [];
	}

	normalize(nodes) {
		return [nodes];
	}
}

function isValidNode(node, selector) {
	return (
		selector &&
		((selector.type === 'tag' && selector.value === node.type) ||
			selector.type === 'universal')
	);
}

var getCombinator = (node, selector) => {
	const container = new PassthroughContainer();
	const result = [];

	const nodeIndex = node.parent.index(node);
	const nextNode = node.next();
	const nextSelector = selector.next();
	const isNodeContainer = node instanceof Container;

	if (selector.value === ' ' && isNodeContainer) {
		node.walk((resolvedNode) => {
			container.append(resolvedNode);
		});
		result.push(container);
	}

	if (selector.value === '+') {
		if (typeof nextNode !== 'undefined') {
			if (isValidNode(nextNode, nextSelector)) {
				container.append(nextNode);
			}
		}
		result.push(container);
	}

	if (selector.value === '~') {
		node.parent.each((resolvedNode, index) => {
			if (index > nodeIndex && isValidNode(resolvedNode, nextSelector)) {
				container.append(resolvedNode);
			}
		});
		result.push(container);
	}

	if (selector.value === '>' && isNodeContainer) {
		node.each((resolvedNode) => {
			if (nextSelector && isValidNode(resolvedNode, nextSelector)) {
				container.append(resolvedNode);
			}
		});
		result.push(container);
	}

	return result;
};

function addNodeToResult(result) {
	return (resolvedNode) => {
		result.push(resolvedNode);
	};
}

var getTag = (node, selector) => {
	const previousSelector = selector.prev();
	const result = [];

	const isPreviousSelectorCombinator =
		previousSelector &&
		previousSelector.type === 'combinator' &&
		['>', '+', '~'].includes(previousSelector.value);
	const callback = addNodeToResult(result);

	if (selector.type === 'universal') {
		if (isPreviousSelectorCombinator) {
			node.each(callback);
		} else {
			node.walk(callback);
		}
	}

	if (selector.type === 'tag') {
		if (isPreviousSelectorCombinator) {
			node.each((resolvedNode) => {
				if (selector.value === resolvedNode.type) {
					callback(resolvedNode);
				}
			});
		} else {
			switch (selector.value) {
				case 'decl':
					node.walkDecls(callback);
					break;
				case 'atrule':
					node.walkAtRules(callback);
					break;
				case 'rule':
					node.walkRules(callback);
					break;
				case 'comment':
					node.walkComments(callback);
					break;
			}
		}
	}

	return result;
};

const REGEX_REGEX = /^\/(.+)\/([gimuy]+)?$/; // eslint-disable-line unicorn/no-unsafe-regex

function isRegexMatched(regex, flags, node, selector) {
	return new RegExp(regex, flags).test(node[selector.attribute]);
}

function whitespaceSeparatedWordTest(selector) {
	return (value) => {
		if (selector.insensitive) {
			return value.toLowerCase() === selector.value.toLowerCase();
		}
		return value === selector.value;
	};
}

function getRegexByOperator(operator, value) {
	switch (operator) {
		case '|=':
			return `^${value}-?`;
		case '^=':
			return `^${value}`;
		case '$=':
			return `${value}$`;
		case '*=':
		default:
			return `${value}`;
	}
}

var getAttribute = (node, selector) => {
	const result = [];
	const hasSelectorAttribute = selector.attribute in node;

	if (
		hasSelectorAttribute &&
		(typeof selector.operator === 'undefined' ||
			typeof selector.value === 'undefined')
	) {
		result.push(node);
	}

	if (
		hasSelectorAttribute &&
		selector.operator === '=' &&
		selector.quoted &&
		REGEX_REGEX.test(selector.value)
	) {
		const [, regex, flags = ''] = selector.value.match(REGEX_REGEX);
		if (isRegexMatched(regex, flags, node, selector)) {
			result.push(node);
		}
	}

	if (hasSelectorAttribute && selector.operator === '=') {
		if (
			(node[selector.attribute] === selector.value &&
				!selector.insensitive) ||
			(node[selector.attribute].toLowerCase() ===
				selector.value.toLowerCase() &&
				selector.insensitive)
		) {
			result.push(node);
		}
	}

	if (hasSelectorAttribute && selector.operator === '~=') {
		if (
			node[selector.attribute]
				.split(' ')
				.some(whitespaceSeparatedWordTest(selector))
		) {
			result.push(node);
		}
	}

	if (
		hasSelectorAttribute &&
		['|=', '^=', '$=', '*='].includes(selector.operator)
	) {
		const flags = selector.insensitive ? 'i' : '';
		const regex = getRegexByOperator(selector.operator, selector.value);

		if (regex !== null && isRegexMatched(regex, flags, node, selector)) {
			result.push(node);
		}
	}

	return result;
};

function processNthNodeQuery(selector, siblings) {
	const [first, , third] = selector.first.map(({ value }) => value);
	const [, negative = '', number] = first.match(NTH_CHILD_QUERY_REGEX) || [];
	const parsedNumber = Number(number);
	const parsedThird = Number(third);
	const range = new Array(
		Number(negative !== '' ? parsedThird : siblings.length)
	)
		.fill(0)
		.map((value, index) => value + index);

	const indices = range
		.map((index) => {
			if (negative !== '') {
				return -(parsedNumber * index) + parsedThird;
			}
			return parsedNumber * index + parsedThird;
		})
		.map((index) => index - 1)
		.filter((index) => index >= 0 && index <= siblings.length - 1);

	const uniqueIndices = [...new Set(indices)];

	return uniqueIndices;
}

var getPseudo = (node, selector, processSelectors) => {
	const result = [];

	if (selector.value === ':empty') {
		if (node.nodes && node.nodes.length === 0) {
			result.push(node);
		}
	}

	if (selector.value === ':only-child') {
		if (
			node.parent.nodes &&
			node.parent.nodes.length === 1 &&
			node.parent.nodes[0] === node
		) {
			result.push(node);
		}
	}

	if (selector.value === ':first-child') {
		if (node.parent.first === node) {
			result.push(node);
		}
	}

	if (selector.value === ':last-child') {
		if (node.parent.last === node) {
			result.push(node);
		}
	}

	if (
		selector.value === ':nth-child' ||
		selector.value === ':nth-last-child'
	) {
		const indices = processNthNodeQuery(selector, node.parent.nodes);
		const nodesCount = node.parent.nodes.length - 1;
		indices.forEach((index) => {
			const nodeIndex =
				selector.value === ':nth-child' ? index : nodesCount - index;
			if (node.parent.nodes[nodeIndex] === node) {
				result.push(node);
			}
		});
	}

	if (
		selector.value === ':first-of-type' ||
		selector.value === ':last-of-type' ||
		selector.value === ':only-of-type' ||
		selector.value === ':nth-of-type' ||
		selector.value === ':nth-last-of-type'
	) {
		const typeResult = [];
		node.parent.each((resolvedNode) => {
			if (resolvedNode.type === node.type) {
				typeResult.push(resolvedNode);
			}
		});
		const nodesCount = typeResult.length;

		typeResult.forEach((resolvedTypeResultNode, index) => {
			if (
				selector.value === ':first-of-type' &&
				index === 0 &&
				resolvedTypeResultNode === node
			) {
				result.push(node);
			}
			if (
				selector.value === ':last-of-type' &&
				index === nodesCount - 1 &&
				resolvedTypeResultNode === node
			) {
				result.push(node);
			}
			if (
				selector.value === ':only-of-type' &&
				nodesCount === 1 &&
				resolvedTypeResultNode === node
			) {
				result.push(node);
			}
		});

		if (
			selector.value === ':nth-of-type' ||
			selector.value === ':nth-last-of-type'
		) {
			const indices = processNthNodeQuery(selector, typeResult);
			const nthNodesCount = typeResult.length - 1;
			indices.forEach((index) => {
				const nodeIndex =
					selector.value === ':nth-of-type'
						? index
						: nthNodesCount - index;
				if (typeResult[nodeIndex] === node) {
					result.push(node);
				}
			});
		}
	}

	if (selector.value === ':not' || selector.value === ':matches') {
		const container = new PassthroughContainer();
		container.append(node);

		const processSelectorsResult = processSelectors(selector, container);

		if (selector.value === ':not') {
			if (!processSelectorsResult.includes(node)) {
				result.push(node);
			}
		} else {
			processSelectorsResult.forEach(
				(resolvedProcessSelectorsResultNode) => {
					result.push(resolvedProcessSelectorsResultNode);
				}
			);
		}
	}

	return result;
};

function processSelectors(selectors, ast) {
	const nodes = selectors
		.map((rootSelector) =>
			rootSelector.reduce(
				(astContainer, selector) =>
					astContainer
						.map((node) => {
							switch (selector.type) {
								case 'combinator':
									return getCombinator(node, selector);

								case 'attribute':
									return getAttribute(node, selector);

								case 'pseudo':
									return getPseudo(
										node,
										selector,
										processSelectors
									);

								case 'tag':
								case 'universal':
								default:
									return getTag(node, selector);
							}
						})
						.reduce((array, result) => [...array, ...result], [])
						.filter((result) => result !== null),
				[ast]
			)
		)
		.reduce((array, result) => [...array, ...result], []);

	const uniqueNodes = [...new Set(nodes)];

	return uniqueNodes;
}

var index = async (query, postcssAst) => {
	const selectorAst = await getSelectorAst(query);
	return processSelectors(selectorAst, postcssAst);
};

export default index;
//# sourceMappingURL=index.esm.js.map
